#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "tetX.h"
#include "tet_values.h"

extern unsigned long BkgrColour;
extern int ScreenWidth, ScreenHeight, ScreenDepth;
extern int screenNumber;

extern Display	*display;
extern Visual   *visual;
extern Window	window;
extern Colormap   colmap;
extern GC  graphContext;
extern Pixmap   screenContext;
extern  Bool DoInvalidate;


static unsigned char StandardPalette[48]=
   {  0,  0,  0,    0,  0, 42,    0, 42,  0,    0, 42, 42,
     42,  0,  0,   42,  0, 42,   42, 21,  0,   42, 42, 42,
     21, 21, 21,   21, 21, 63,   21, 63, 21,   21, 63, 63,
     63, 21, 21,   63, 21, 63,   63, 63, 21,   63, 63, 63
   };

int	PalColours;
//unsigned long *PixValues = NULL;
unsigned long PixValues[16];
extern char *BgColourName;
unsigned long BkgrXColour;


static unsigned long AllocRGBColour(unsigned short r, unsigned short g, 
                             unsigned short b)
{   XColor  xc;

    xc.red = r << 10;
    xc.green = g << 10;
    xc.blue = b << 10;
    xc.flags = DoRed | DoGreen | DoBlue;
    xc.pad = 0;

   if (XAllocColor(display, colmap, &xc)) return xc.pixel;
   
   return BlackPixel(display, screenNumber);
}   

unsigned long AllocColour(char *ColourName)
{   XColor  xc;
    long unsigned result;

    if (!ColourName) goto TakeBlack;
        if (ColourName[0] == '0')
    {  sscanf(ColourName, "%lx", &result);
       xc.blue = (result & 0xFF) << 8;
       xc.green = (result & 0xFF00);
       xc.red =   (result & 0xFF0000) >> 8;
       xc.flags = DoRed | DoGreen | DoBlue;
       xc.pad = 0;
    }
    else
    {  XColor xc1;
       XLookupColor(display, colmap, ColourName, &xc, &xc1); 
    }
    
    if (XAllocColor(display, colmap, &xc) == 0)
    {  fprintf (stderr, "Failed to allocate colour %s, black colour will be used\n", ColourName);
       goto TakeBlack;
    }
    
    return xc.pixel;

 TakeBlack:
    return  AllocRGBColour(0,0,0);  // BlackPixel(display, screenNumber);               
}    


int initPalette(void)
{   int i, j;
    unsigned short r,g,b;

    if (BgColourName == NULL ||  BgColourName[0] == '\0')
	BkgrXColour = AllocRGBColour(0,0,0); // XBlackPixel(display, screenNumber);
    else
        BkgrXColour = AllocColour(BgColourName);   

    PixValues[0] = BkgrXColour;	

    for(i=1,j=3; i<16; i++)
    { r =  StandardPalette[j++];
      g =  StandardPalette[j++];
      b =  StandardPalette[j++];

      PixValues[i] = AllocRGBColour(r, g, b);
    }  

    return 0;
}

void deinitPalette(void)
{
    return;
}


void   adjustPalette(IMAGE_HEADER *imHdr, int maxcols)
{

 /*  This function will allocate colours for a 256c picture in 16 mode */
 
   int used = imHdr->usedcol;
   int PalColsUsed = 16;
   PAL_ENTRY curImPal;
   PAL_ENTRY *ImPal = imHdr->pal;
   COLOUR_REF *PalEquiv = imHdr->PalEqu;
   long dist, sdist;
   short dr, dg, db;
   short i, j, sind;

#ifndef SHOW_256C_IMAGE
   if (ScreenDepth <= 4 && imHdr->imbpp > 4) return;
#endif


   for (i=0; i<used; i++)
   {
      curImPal = ImPal[i];
      sdist = MAXLONG;
      sind = -1;
      for (j=0; j<PalColsUsed; j++)
      {
         dr = StandardPalette[j*3] - curImPal.r;
         dg = StandardPalette[j*3+1] - curImPal.g;
         db = StandardPalette[j*3+2] - curImPal.b;
         dist = dr*dr + dg*dg + db*db;
         if (dist < sdist)
         { sind = j; sdist = dist;
           if (sdist == 0) break;
         }
      }

      dr = StandardPalette[sind*3];
      dg = StandardPalette[sind*3+1];
      db = StandardPalette[sind*3+2];
      PalEquiv[i] = AllocRGBColour(dr,dg,db);
      
    }


}

void   setPalette(IMAGE_HEADER *imHdr, int maxcols)
{

     int used = imHdr->usedcol;
     PAL_ENTRY *ImPal = imHdr->pal;
     PAL_ENTRY curImPal;
     COLOUR_REF  *PalEquiv = imHdr->PalEqu;
     int	i;

     if (ScreenDepth <= 4)
     {  adjustPalette(imHdr, maxcols); return; }
	    
     if (ScreenDepth <= 8)
     {  int colours = (1 << ScreenDepth);
        if (used  > colours) used = colours;
     }
     	
     for (i=0; i<used; i++)
     {
        curImPal = ImPal[i];
        PalEquiv[i] =
	     AllocRGBColour(curImPal.r, curImPal.g, curImPal.b);
     }
}     	     




int   showSqueezedImage(IMAGE_HEADER *imHdr, int x, int y, int transp,
                                               int stepx, int stepy)
{
   XImage  *xim;    

   short wid = imHdr->imwid;
   short hei = imHdr->imhei;
   short bpp = imHdr->imbpp;
   int   x1, y1;
   unsigned char *imstart = imHdr->image, *imptr;
   COLOUR_REF *PalEquiv = imHdr->PalEqu;
   short c, col;
   int  reswid, reshei;
   short i, j, k, towid;

#ifndef SHOW_256C_IMAGE
   if (ScreenColours <= 16 && bpp > 4) return 0;
#endif

   reswid = (wid+stepx-1)/stepx; 
   reshei = (hei+stepy-1)/stepy;
  
   xim = XGetImage(display, screenContext, x, y, 
                             reswid, reshei, -1, ZPixmap);
   if (!xim) return 8;			     

   towid = wid;
   if (bpp <= 4) towid >>= 1;

   y1 = 0;      
   for (i=0; i<hei; i+=stepy)
   { 
       imptr = imstart + i*towid;
       x1 = 0;
       for (j=0; j<towid; j+=stepx)
       {  c = '\0';
          for (k=0; k<stepx; k++)  c |= *imptr++;
          if (bpp > 4) col=c;
          else
          {  col = c >> 4;
             if (col != transp)
                     XPutPixel(xim, x1, y1, PalEquiv[col]); 
             x1 ++;
             col = c & 0xf;
          }

          if (col != transp)
              XPutPixel(xim, x1, y1, PalEquiv[col]); 
          x1++;
       }

       y1++;    
  }

    XPutImage(display, screenContext, graphContext, xim,
              0, 0, x, y, reswid, reshei);

    if (DoInvalidate)
    { RECT rect;
      rect.top = y; rect.left = x;
      rect.bottom = y+reshei; rect.right = x + reswid;
      ContextRectToScreen(&rect);
    }  	      
	      
    XDestroyImage(xim);	      
    return 0;
}

int   showImage(IMAGE_HEADER *imHdr, int x, int y, int transp)
{
  return showSqueezedImage(imHdr, x, y, transp, 1, 1);
}


