/*********************************************************/
/*  This source code is copyrighted by the author        */
/*  Please read COPYING  for the GNU licence policy.     */
/*********************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "tetX.h"
#include "egafont.h"

extern  Display *display;
extern  Pixmap  screenContext;
extern  Window  window;
extern  int	ScreenDepth;

extern  GC      graphContext;
extern  Bool    DoInvalidate;
extern COLOUR_REF  PixValues[];

static XFontStruct *FontData; 
static int  FontHeight;

#ifdef TET_USE_WIDE_CHARS
int wide_chars = 1;

int mbstoXChar2b(XChar2b **rtarget, char *source)
{
   wchar_t *textbuf=NULL, *ptextbuf, textchar;
   XChar2b *target=NULL, *ptarget;
   int   len;  /* String length*/
   int   lenPlus;  /* String length + 1*/

   len= mbstowcs(NULL, source, 0);
   lenPlus = len + 1;
   textbuf = malloc(lenPlus * sizeof(wchar_t));
   target  = malloc(lenPlus * sizeof(XChar2b));
   if (textbuf == NULL || target == NULL)
   { len = -1;  goto StopIt; }

   mbstowcs(textbuf, source, lenPlus);
   ptextbuf = textbuf;
   ptarget = target;
   while (--lenPlus >= 0)
   {
	  textchar = *ptextbuf++;
	  ptarget->byte1 = textchar >> 8;
	  ptarget->byte2 = textchar;
	  ptarget++;
   }       	

StopIt:
   if (textbuf) free(textbuf);

   if (len < 0 && target) 
   {  free(target); target = NULL; }

   *rtarget = target;
   return len;
}


#else
int	wide_chars = 0;
#endif


int egaShowText(char *text, int x, int y, int fgcol, int bgcol)
{  XGCValues xgcv;
   int  len, wid;

   if (!FontData) return 1;

   xgcv.foreground = PixValues[fgcol];
   xgcv.background = PixValues[bgcol];

   XChangeGC(display, graphContext, GCForeground | GCBackground ,
                                     &xgcv);

#ifdef TET_USE_WIDE_CHARS
  {	XChar2b *widebuf;  
	len = mbstoXChar2b(&widebuf, text);
	if (len < 0) return 1;
	XDrawImageString16 (display, screenContext, graphContext,
                        x, y+FontData->ascent, widebuf, len);
    wid = XTextWidth16(FontData, widebuf, len);
	free(widebuf);
  }	 					 
#else   
   len  = strlen(text);
   XDrawImageString (display, screenContext, graphContext,
                    x, y+FontData->ascent, text, len);
   wid = XTextWidth(FontData, text, len);
#endif


   if (DoInvalidate)
   { RECT rect;

     rect.left = x; 
     rect.top = y; //FontData->descent;
     rect.bottom = y + FontHeight;
     rect.right = rect.left + wid;
     ContextRectToScreen(&rect);
   }     

   return 0;

}

int egaShowScaledText(char *text, int x, int y,
                      int fgcol, int bgcol, int hscale, int vscale)
{
    int wid, hei, thei, twid;
    int len;
    XGCValues xgcv;
    XImage   *xim;
    int      x1, y1, x2, y2, x3, y3, i, j;
    COLOUR_REF pixel;
     
    if (!FontData) return 1;

    xgcv.foreground = PixValues[fgcol];
    xgcv.background = PixValues[bgcol];
  
    XChangeGC(display, graphContext, GCForeground | GCBackground ,
                                     &xgcv);

#ifdef TET_USE_WIDE_CHARS
    { XChar2b *widebuf;  
	  len = mbstoXChar2b(&widebuf, text);
	  if (len < 0) return 1;
	  XDrawImageString16 (display, screenContext, graphContext,
                        x, y+FontData->ascent, widebuf, len);
	  wid = XTextWidth16(FontData, widebuf, len);
	  free(widebuf);
    }	 					 
#else   
	len  = strlen(text);
	XDrawImageString (display, screenContext, graphContext,
                    x, y+FontData->ascent, text, len);
	wid = XTextWidth(FontData, text, len);
#endif

    hei = FontHeight;
    twid = wid * hscale;
    thei = hei * vscale;
    
    xim = XGetImage(display, screenContext, x, y,
                          twid, thei, AllPlanes, ZPixmap);

    if (!xim) return 2;
        
    /* Now scale! */
    for (y3=hei-1; y3>=0; y3--)
    {  y1 = y3*vscale;

       // From a single line
       for (x3=wid-1; x3>=0; x3--)
       {  pixel = XGetPixel(xim, x3, y3);

	  	  // Spread to whole square          
		  x1 = x3 * hscale;
		  y2 = y1;
		  for (j = 0; j < vscale; j++)
		  {   x2 = x1;
	  		  for (i=0; i < hscale; i++)
	      	   XPutPixel(xim, x2++, y2, pixel);
		      y2++;	 
		  }      
	  
       }
    }   
             	       	       

    XPutImage(display, screenContext, graphContext, xim, 0, 0,
                       x, y, twid, thei);

    XDestroyImage(xim);
            
   if (DoInvalidate)
   { RECT rect;

     rect.left = x; 
     rect.top = y; //FontData->descent;
     rect.bottom = y + thei;
     rect.right = rect.left + twid;
     ContextRectToScreen(&rect);
   }     

   return 0;
    

}

int egaSetFont(char *cFontData)
{   
    FontData = (XFontStruct *) cFontData;

    if (FontData)
    {    XSetFont(display, graphContext, FontData->fid);
	 FontHeight = FontData->ascent + FontData->descent;
    }	 
    else
        FontHeight = 0;
    
   return 0;
}

void egaFreeFont(char *cFont)
{
   if (cFont)
      XFreeFont(display, (XFontStruct *) cFont);
}


int egaGetFontHeight(void)
{
     return  FontHeight;
}

int egaGetTextWidth(char *text)
{
  int len, wid;
  
  if (!FontData) return 0;

#ifdef TET_USE_WIDE_CHARS
  {	tchar *textbuf;  
	len = mbstowcs(NULL, text, 0);
    textbuf = malloc((len+1) * sizeof(wchar_t));
	if (textbuf == NULL) return 1;
    mbstowcs(textbuf, text, len+1);
	wid = XTextWidth16(FontData, (XChar2b *) textbuf, len);
	free(textbuf);
  }	 					 
#else   
  len  = strlen(text);
  wid = XTextWidth(FontData, text, len);
#endif

  return wid;
}

FONT_TYPE egaFontType(void)
{
   return FONT_TYPE_X11_NATURAL;
}

/*
char *egaGetFontPtr(int chr)
{
    return (char *) FontData;   // ???? not for a partic char
}
*/

Bool GetFontInfoAndCheck(XFontStruct *xfs, char *target)
{
    char *Foundry = NULL;
    char *Charset = NULL;
    char *Weight = NULL;
    unsigned long   wid, hei;
    Atom atm;
    unsigned long value; 
    Bool	status;
/*    
    int  i;
    
    for (i=0; i<xfs->n_properties; i++)
    {   atm = (xfs->properties[i]).name;        // Atom
	value = (xfs->properties[i]).card32;    
        printf ("%ld. %s = %ld\n", atm, XGetAtomName(display, atm), value);
    }
*/
    wid = 8;
    status = False;
    atm = XInternAtom(display, "AVERAGE_WIDTH", True);
    if (atm != None && XGetFontProperty(xfs, atm, &value))
               wid =  value/10;

/*    if (wid != 8) goto Ret; */
        
    hei = 13;
    atm = XInternAtom(display, "PIXEL_SIZE", True);
    if (atm != None && XGetFontProperty(xfs, atm, &value))
               hei =  value;
    if (hei <10 || hei > 16) goto Ret;

    Foundry = "";
    atm = XInternAtom(display, "FOUNDRY", True);
    if (atm != None && XGetFontProperty(xfs, atm, &value))
               Foundry = XGetAtomName(display, (Atom)value);
	       
    Weight = "";
    atm = XInternAtom(display, "WEIGHT_NAME", True);
    if (atm != None && XGetFontProperty(xfs, atm, &value))
               Weight = XGetAtomName(display, (Atom)value);
	       
    Charset = "";
    atm = XInternAtom(display, "CHARSET_REGISTRY", True);
    if (atm != None && XGetFontProperty(xfs, atm, &value))
               Charset = XGetAtomName(display, (Atom)value);
	 
    
    sprintf (target, "%s %ldx%ld %s (%s)", Foundry, wid, hei,
                                 Weight, Charset);
				 
    status = True;				 

Ret:

    if (Foundry) XFree(Foundry);
    if (Charset) XFree(Charset);
    if (Weight) XFree(Weight);
    return status;    
    
}

char *egaLoadFont(char *fontName, char *fontDir, char *message)
{
    char *DefaultFonts[] =
        { NULL,
	  "-*-fixed-bold-r-*-*-16-*-*-*-*-80-*-*",
	  "-*-fixed-bold-r-*-*-13-*-*-*-*-80-*-*",
	  "-*-fixed-*-r-*-*-16-*-*-*-*-80-*-*",
	  "-*-fixed-*-r-*-*-13-*-*-*-*-80-*-*",
          "fixed", NULL 
	};
    char *fntNameCur;
    char **fntNamePtr;
    XFontStruct  *xfs;

    fntNamePtr = DefaultFonts;
    if (fontName)  DefaultFonts[0] = fontName; 
    else   fntNamePtr++;
    
    while ((fntNameCur = *fntNamePtr++) != NULL)
    {   xfs = XLoadQueryFont( display,  fntNameCur);
        if (xfs && GetFontInfoAndCheck(xfs, fontName))
	   return (char *)xfs;
    }
    
    strcpy(message, "Could not find appropriate fonts");
	
    return (char *) NULL;
    	   
}   


int egaShowCharSpecial(tchar textChar, int x, int y, int fgcol, int bgcol,
                       int hscale, int vscale, int flags)
{  

    int wid, hei, thei, twid;
    XGCValues xgcv;
    XImage   *xim;
    int      x1, y1, x2, y2, x3, y3, i, j;
    COLOUR_REF pixel, pixel1;
     
    if (!FontData) return 1;

    xgcv.foreground = PixValues[fgcol];
    xgcv.background = PixValues[bgcol];
  
    XChangeGC(display, graphContext, GCForeground | GCBackground ,
                                     &xgcv);

#ifdef TET_USE_WIDE_CHARS
    { XChar2b widechar;
	  widechar.byte1 = textChar >> 8;
	  widechar.byte2 = textChar;
	  wid = XTextWidth16(FontData, &widechar, 1);
      XDrawImageString16 (display, screenContext, graphContext,
                    x, y+FontData->ascent, &widechar, 1);
    }					
#else
    wid = XTextWidth(FontData, &textChar, 1);
    XDrawImageString (display, screenContext, graphContext,
                    x, y+FontData->ascent, &textChar, 1);
#endif	
    hei = FontHeight;
    twid = wid * hscale;
    thei = hei * vscale;



    xim = XGetImage(display, screenContext, x, y,
                          twid, thei, AllPlanes, ZPixmap);

    if (!xim) return 2;


    /* flip */
    y2 = hei >> 1;
    x2 = wid >> 1;
    if (flags & EGAFONT_HFLIP)
    {   
       for (y3=0; y3< hei; y3++)
         for (x3=0, x1=wid-1; x3 < x2; x3++, x1--)       
         {  pixel = XGetPixel(xim, x3, y3);
	    pixel1 = XGetPixel(xim, x1, y3);
            XPutPixel(xim, x3, y3, pixel1);
            XPutPixel(xim, x1, y3, pixel);
         }
    }	 	    
        
    if (flags & EGAFONT_VFLIP)
    {   
       for (x3=0; x3< wid; x3++)
         for (y3=0, y1=hei-1; y3 < y2; y3++, y1--)       
         {  pixel = XGetPixel(xim, x3, y3);
	    pixel1 = XGetPixel(xim, x3, y1);
            XPutPixel(xim, x3, y3, pixel1);
            XPutPixel(xim, x3, y1, pixel);
         }
    }	 	    

    /* Now scale! */
    for (y3=hei-1; y3>=0; y3--)
    {  y1 = y3*vscale;

       // From a single line
       for (x3=wid-1; x3>=0; x3--)
       {  pixel = XGetPixel(xim, x3, y3);

	  
	  // Spread to whole square          
	  x1 = x3 * hscale;
	  y2 = y1;
	  for (j = 0; j < vscale; j++)
	  {   x2 = x1;
	      for (i=0; i < hscale; i++)
	         XPutPixel(xim, x2++, y2, pixel);
	      y2++;	 
	  }      
	  
       }
    }   
             	       	       

    XPutImage(display, screenContext, graphContext, xim, 0, 0,
                       x, y, twid, thei);

    XDestroyImage(xim);
            
   if (DoInvalidate)
   { RECT rect;

     rect.left = x; 
     rect.top = y; //FontData->descent;
     rect.bottom = y + thei;
     rect.right = rect.left + twid;
     ContextRectToScreen(&rect);
   }     

   return 0;
    

}

XImage *DrawCharOnXImage(tchar textChar, unsigned long fgrPix, unsigned long bgrPix)
{  
   Pixmap tmpPixmap;    
   XImage *xim;    
   XGCValues xgcv;    
   int   wid, hei;
#ifdef TET_USE_WIDE_CHARS
   XChar2b widechar;
#endif	

   if (!FontData) return (XImage *)NULL;

#ifdef TET_USE_WIDE_CHARS
    widechar.byte1 = textChar >> 8;
    widechar.byte2 = textChar;
    wid = XTextWidth16(FontData, &widechar, 1);
#else
   wid = XTextWidth(FontData, &textChar, 1);
#endif   

   hei = FontHeight;
   
   tmpPixmap = XCreatePixmap(display, window, wid, hei, ScreenDepth);
   if (tmpPixmap == None) return (XImage *)NULL;


   xgcv.foreground = fgrPix;
   xgcv.background = bgrPix;
  
   XChangeGC(display, graphContext, GCForeground | GCBackground ,
                                     &xgcv);


#ifdef TET_USE_WIDE_CHARS
   XDrawImageString16 (display, tmpPixmap, graphContext,
                    0, FontData->ascent, &widechar, 1);
#else
   XDrawImageString (display, tmpPixmap, graphContext,
                    0, FontData->ascent, &textChar, 1);
#endif

   xim = XGetImage(display, tmpPixmap, 0, 0, wid, hei, AllPlanes, ZPixmap);

   XFreePixmap(display, tmpPixmap);
   
   return xim;   
}

/*
XImage *DrawCharOnXImage(tchar textChar, unsigned long fgrPix, unsigned long bgrPix)
{  
   Pixmap tmpPixmap;    
   XImage *xim;    
   XGCValues xgcv;    
   int   wid, hei;
   char txtchar = textChar & 0xff;
#ifdef TET_USE_WIDE_CHARS
   XChar2b widechar;
#endif	

   if (!FontData) return (XImage *)NULL;

#ifdef TET_USE_WIDE_CHARS
    widechar.byte1 = textChar >> 8;
    widechar.byte2 = textChar;
    wid = XTextWidth16(FontData, &widechar, 1);
#else
//    wid = XTextWidth(FontData, &textChar, 1);
    wid = XTextWidth(FontData, &txtchar, 1);
#endif	

   xgcv.foreground = fgrPix;
   xgcv.background = bgrPix;
  
   XChangeGC(display, graphContext, GCForeground | GCBackground ,
                                     &xgcv);

   hei = FontHeight;
   
   tmpPixmap = XCreatePixmap(display, window, wid, hei, ScreenDepth);
   if (tmpPixmap == None) return (XImage *)NULL;

#ifdef TET_USE_WIDE_CHARS
   XDrawImageString16 (display, screenContext, graphContext,
                  0, FontData->ascent, &widechar, 1);
#else
//   XDrawImageString (display, screenContext, graphContext,
//                  0, FontData->ascent, &textChar, 1);
   XDrawImageString (display, screenContext, graphContext,
                  0, FontData->ascent, &txtchar, 1);
#endif	

   xim = XGetImage(display, tmpPixmap, 0, 0, wid, hei, AllPlanes, ZPixmap);

   XFreePixmap(display, tmpPixmap);
   
   return xim;   
}
*/

